import os
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from data_process import load_mat, get_shuffle_mat
from get_bds import get_bds, select_bds
from dp import get_dp_mat, get_tads
from quant_reg import quant_reg
import numpy as np
import time
import argparse
from get_stripe import get_stripe


def JOnTADS_one(file_name, output, max_sz, min_sz, significance, percentile, num_ft_cuts, lmda):
    print('Load contact matrix and shuffle it, spend time: ')
    start = time.time()
    mat, cum_mat = load_mat(file_name, max_sz)
    mat_shuffle, cum_mat_shuffle = get_shuffle_mat(mat, max_sz)
    end = time.time()
    print('%.2fs' % (end-start))
    
    # identify boundaries
    print('Identify boundaries, spend time: ')
    start = time.time()
    bds = get_bds(mat, mat_shuffle, max_sz, min_sz, percentile, num_ft_cuts)
    bds_shuffle = get_bds(mat_shuffle, mat_shuffle, max_sz, min_sz, percentile, num_ft_cuts)
    end = time.time()
    print('%.2fs' % (end - start))

    # get tads
    print('Get tads, spend time: ')
    start = time.time()
    dp_mat_shuffle = get_dp_mat(np.array(bds_shuffle), mat_shuffle, cum_mat_shuffle, max_sz, min_sz, np.zeros(max_sz-min_sz+1))
    tads_shuffle = get_tads(bds_shuffle, dp_mat_shuffle, max_sz, min_sz)
    dp_mat = get_dp_mat(np.array(bds), mat, cum_mat, max_sz, min_sz, np.zeros(max_sz-min_sz+1))
    tads = get_tads(bds, dp_mat, max_sz, min_sz)
    end = time.time()
    print('%.2fs' % (end - start))
    
    # filter tads and boundaries
    print('Filter tads and boundaries, spend time: ')
    start = time.time()
    idx, curve = quant_reg(tads, tads_shuffle, max_sz, min_sz, significance, lmda)
    bds = np.unique(tads[idx][:, :2])
    end = time.time()
    print('%.2fs' % (end - start))

    # use filtered boundaries to identify tads
    print('Use filtered boundaries to identify tads, spend time: ')
    start = time.time()
    dp_mat = get_dp_mat(np.array(bds, dtype=int), mat, cum_mat, max_sz, min_sz, np.zeros(max_sz-min_sz+1))
    tads = get_tads(bds, dp_mat, max_sz, min_sz)
    end = time.time()
    print('%.2fs' % (end - start))

    # save tads
    print('Save tads in the output file.')
    np.savetxt(output, tads, fmt='%d, %d, %.3g, %d')

def JOnTADS(file_names, outputs, max_sz, min_sz, significance, percentile, num_ft_cuts, lmda):
    if len(file_names) == 1:
        JOnTADS_one(file_names[0], outputs[0], max_sz, min_sz, significance, percentile, num_ft_cuts, lmda)
        return 1
    
    def declutter_bds(bds_all, mats, max_sz, min_sz):
        max_num = 0
        idx_max_num = -1
        bds_num = []
        for i in range(len(bds_all)):
            bds_num.append(len(bds_all[i]))
            if len(bds_all[i]) > max_num:
                max_num = len(bds_all[i])
                idx_max_num = i
        ratio = []
        for i in range(len(bds_all)):
            ratio.append(len(bds_all[i])/max_num)
        ft_all = []
        for i in range(len(mats)):
            ft = []
            for j in range(mats[i].shape[0]):
                ft.append(np.mean(np.sort(mats[i][j, max(0, j-max_sz+1):min(mats[i].shape[0], j+max_sz)])[-min_sz:]))
            ft_all.append(ft)
        ft_all = np.array(ft_all)
        ft_pooled = np.mean(ft_all, axis=0)
        bds_all = np.unique([j for i in bds_all for j in i])
        bds_all = [np.array(select_bds(bds_all, ft_pooled, int((min_sz+1)/2))) for i in range(len(mats))]
        bds_all = [np.sort(np.array(bds_all[i])[np.argsort(ft_all[i][bds_all[i]])[int(-len(bds_all[idx_max_num])*ratio[i]):]]) for i in range(len(mats))]
        return bds_all
        
    print('Load contact matrices and shuffle them, spend time: ')
    start = time.time()
    mats, cum_mats, mats_shuffle, cum_mats_shuffle = [], [], [], []   
    for file_name in file_names:
        _mat, _cum_mat = load_mat(file_name, max_sz)
        _mat_shuffle, _cum_mat_shuffle = get_shuffle_mat(_mat, max_sz)
        mats.append(_mat)
        cum_mats.append(_cum_mat)
        mats_shuffle.append(_mat_shuffle)
        cum_mats_shuffle.append(_cum_mat_shuffle)
    end = time.time()
    print('%.2fs' % (end - start))

    print('Get boundaries, spend time: ')
    start = time.time()
    bds_each = []
    bds_all, bds_all_shuffle = [], []  
    bds_vas = []
    for mat, mat_shuffle in zip(mats, mats_shuffle):
        _bds = get_bds(mat, mat_shuffle, max_sz, min_sz, percentile, num_ft_cuts)
        bds_all += _bds
        bds_each.append(_bds)
        bds_all_shuffle.append(get_bds(mat_shuffle, mat_shuffle, max_sz, min_sz, percentile, num_ft_cuts))
    bds_all = declutter_bds(bds_each, mats, max_sz, min_sz)
    end = time.time()
    print('%.2fs' % (end - start))
    
    print('Get tads, spend time: ')
    start = time.time()
    tads_all = []
    for i in range(len(file_names)):
        _dp_mat_shuffle = get_dp_mat(np.array(bds_all_shuffle[i]), mats_shuffle[i], cum_mats_shuffle[i], max_sz, min_sz, np.zeros(max_sz-min_sz+1))
        _tads_shuffle = get_tads(bds_all_shuffle[i], _dp_mat_shuffle, max_sz, min_sz)
        _dp_mat = get_dp_mat(np.array(bds_all[i]), mats[i], cum_mats[i], max_sz, min_sz, np.zeros(max_sz-min_sz+1))
        _tads = get_tads(bds_all[i], _dp_mat, max_sz, min_sz)
        _idx, curve = quant_reg(_tads, _tads_shuffle, max_sz, min_sz, significance, lmda)
        bds_all[i] = np.unique(_tads[_idx][:, :2])

        _dp_mat = get_dp_mat(np.array(bds_all[i], dtype=int), mats[i], cum_mats[i], max_sz, min_sz, np.zeros(max_sz-min_sz+1))
        _tads = get_tads(bds_all[i], _dp_mat, max_sz, min_sz)
        tads_all.append(_tads)
    end = time.time()
    print('%.2fs' % (end - start))
        
    for i in range(len(outputs)):
        np.savetxt(outputs[i], tads_all[i], fmt='%d, %d, %.3g, %d')
    
def get_args():
    parser = argparse.ArgumentParser(description='Jointly identify nested TADs.')
    parser.add_argument('-F', '--file_name', type=str, nargs='+',
                        help='Input file path')
    parser.add_argument('-O', '--output', type=str, nargs='+',
                        help='Output file path')
    parser.add_argument('-MAXSZ', '--max_sz', type=int, default=200,
                        help='Maximum size of TADs')
    parser.add_argument('-MINSZ', '--min_sz', type=int, default=7,
                        help='Minimum size of TADs')
    parser.add_argument('--significance', type=float, default=0.1,
                        help='Significance for filtering TADs and boundaries')
    parser.add_argument('--percentile', type=float, default=90,
                        help='Percentile of boundaries considered')
    parser.add_argument('--num_ft_cuts', type=int, default=5,
                        help='Number of feature cut points')
    parser.add_argument('--num_rg_cuts', type=int, default=5,
                        help='Number of regression cut points')
    parser.add_argument('--seed', type=int, default=0,
                        help='Initialize the random number generator')
    parser.add_argument('--lmda', type=float, default=0.001,
                        help='Lambda of L1 norm in nonparametric quantile regression')

    parser.add_argument('--stripe', type=bool, default=False,
                        help='Call stripes or not')
    parser.add_argument('--stripe_output', type=str, default='',
                        help='Output file to save stripe results')
    parser.add_argument('--chr', type=str, default='',
                        help='Chromosome information for calling stripes')
    parser.add_argument('--p', type=float, default=1,
                        help='Exponent for measuring similarity')
    parser.add_argument('--threshold', type=float, default=0.1,
                        help='Threshold to call stripes')
    args = parser.parse_args()
    return args

def main():
    args = get_args()
    np.random.seed(args.seed)
    if args.stripe == True:
        stripes = get_stripe(args.file_name[0], max_sz=args.max_sz, min_sz=args.min_sz, p=args.p, threshold=args.threshold, mean_threshold=args.threshold)
        res = []
        for stripe in stripes:
            if len(stripe[0]) == 1:
                res.append(['chr'+args.chr, stripe[0][0], stripe[0][0], 'chr'+args.chr, stripe[1][0], stripe[1][1]])
            else:
                res.append(['chr'+args.chr, stripe[0][0], stripe[0][1], 'chr'+args.chr, stripe[1][0], stripe[1][0]])
        np.savetxt(args.stripe_output, res, fmt='%s', delimiter='\t')
    np.random.seed(args.seed)
    JOnTADS(args.file_name, args.output,
               args.max_sz, args.min_sz,
               args.significance, args.percentile,
               args.num_ft_cuts, args.lmda)
    return

if __name__ == '__main__':
    main()



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from qpsolvers import solve_qp
import scipy
from scipy.spatial.distance import pdist, squareform

def kernel(x1, x2, sigma):
    return np.exp((-((x1-x2)**2))/(2*sigma**2))
def d1kernel(x1, x2, sigma):
    return (x2-x1)/sigma**2*kernel(x1, x2, sigma)
def nqr(x, y, tau, sigma, penalty):
    K = np.exp(-squareform(pdist(np.expand_dims(x, axis=1), 'euclidean')) ** 2 / (2 * sigma ** 2))
    D2K = (np.tile(x, (len(x), 1)).T-x)/sigma**2*K
    D1K = D2K.T
    D1D2K = (1/sigma**2-(np.tile(x, (len(x), 1)).T-x)**2/sigma**4)*K
    P = np.vstack((np.hstack((K,D1K)), np.hstack((D2K,D1D2K))))
    P += 1e-6*np.eye(2*len(x))
    q = np.hstack((-y, np.zeros(len(x))))
    C = 1/(penalty*len(x))
    lb = np.hstack((np.ones(len(x))*C*(tau-1), np.zeros(len(x))))
    ub = np.hstack((np.ones(len(x))*C*tau, np.ones(len(x))*np.Inf))
    A = np.hstack((np.ones(len(x)), np.zeros(len(x))))
    b = np.zeros(1)
    res = solve_qp(P=P, q=q, A=A, b=b, lb=lb, ub=ub, solver="quadprog")
    alpha = res[:len(x)]
    beta = res[len(x):]
    for i in range(len(x)):
        if np.abs(alpha[i]-C*(tau-1))>1e-2*C*tau and np.abs(alpha[i]-C*tau)>1e-2*C*tau:
            b = y[i] - sum(alpha*K[i, :] + beta*D1K[i, :])
    return alpha, beta, b
def force_decreasing(curve):
    curve[curve<0] = 0
    for i in range(1, len(curve)):
        if curve[i] > curve[i-1]:
            curve[i] = curve[i-1]
    return curve
def _quant_reg(x, y, max_sz, min_sz, tau, lmda):
    fitted = []
    sigma = (max_sz-min_sz)/10
    alpha, beta, b = nqr(x, y, tau, sigma, lmda)
    for i in range(min_sz, max_sz+1):
        fitted.append(np.sum(alpha*kernel(i, x, sigma)+beta*d1kernel(i, x, sigma))+b)
    return force_decreasing(np.array(fitted))

def quant_reg(tads, tads_shuffle, max_sz, min_sz, significance, lmda):
    def helper(tads_shuffle, min_sz, max_sz):
        ps = np.percentile(tads_shuffle[:, 3], [60, 90])
        idx1 = np.random.choice(np.arange(len(tads_shuffle))[tads_shuffle[:, 3]<=ps[0]], 100)
        idx2 = np.random.choice(np.arange(len(tads_shuffle))[np.array(tads_shuffle[:, 3]>ps[0]) & np.array(tads_shuffle[:, 3]<ps[1])], 100)
        idx3 = np.random.choice(np.arange(len(tads_shuffle))[tads_shuffle[:, 3]>=ps[1]], 100)
        idx4 = []
        #p = (1-50/len(tads_shuffle))*100.
        for i in range(min_sz, max_sz+1):
            scores = tads_shuffle[:, 2][tads_shuffle[:, 3]==i]
            if len(scores) != 0:
                # idx4 += (np.arange(len(tads_shuffle))[np.array(tads_shuffle[:, 2]>np.percentile(scores, p)) & np.array(tads_shuffle[:, 3]==i)]).tolist()
                idx4 += (np.arange(len(tads_shuffle))[np.array(tads_shuffle[:, 2]==np.max(scores)) & np.array(tads_shuffle[:, 3]==i)]).tolist()
        idx4 = np.array(idx4)
        idx = np.array(idx1.tolist()+idx2.tolist()+idx3.tolist()+idx4.tolist())
        return tads_shuffle[:, 3][idx], tads_shuffle[:, 2][idx]
    
    size_p, score_p = helper(tads_shuffle, min_sz, max_sz)
    
    left = 0.5
    right = 1
    diff = 1
    idx_out, curve_out = [], []
    for _ in range(10):
        mid = (left+right)/2
        curve = _quant_reg(size_p, score_p, max_sz, min_sz, mid, lmda)
        if (np.max(curve)-np.min(curve)) > (np.std(tads_shuffle[:, 2]) * 0.15):
            num, idx = above_curve(tads, curve, min_sz)
            num_shuffle, _ = above_curve(tads_shuffle, curve, min_sz)
        else:
            while True:
                print('!')
                size_p, score_p = helper(tads_shuffle, min_sz, max_sz)
                curve = _quant_reg(size_p, score_p, max_sz, min_sz, mid+np.random.uniform(-0.0005, 0.0005, 1)[0], lmda)
                if (np.max(curve)-np.min(curve)) > (np.std(tads_shuffle[:, 2]) * 0.15):
                    num, idx = above_curve(tads, curve, min_sz)
                    num_shuffle, _ = above_curve(tads_shuffle, curve, min_sz)
                    break
        if num_shuffle / (num_shuffle+num) > significance:
            left = mid
        elif num_shuffle / (num+num_shuffle) < significance:
            right = mid
        else:
            #plt.plot(tads[:, 3], tads[:, 2], '.')
            #plt.plot(np.arange(min_sz, max_sz+1), curve)
            #import datetime
            #time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            #plt.show()
            #plt.close()
            return idx, curve
        if np.abs(num_shuffle / (num_shuffle+num) - significance) < diff:
            diff = np.abs(num_shuffle / (num_shuffle+num) - significance)
            idx_out = idx
            curve_out = curve

    #plt.plot(tads[:, 3], tads[:, 2], '.')
    #plt.plot(np.arange(min_sz, max_sz+1), curve_out)
    #import datetime
    #time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    #plt.show()
    #plt.close()
    if left == 0.1 or right == 1:
        print('TADs might not exist.')
    return idx_out, curve_out
    
def above_curve(tads, curve, min_sz):
    size = tads[:, 3]
    score = tads[:, 2]
    ans = 0
    idx = []
    for i, sz in enumerate(np.array(size, dtype=int)):
        if score[i] >= curve[sz-min_sz]:
            ans += 1
            idx.append(i)
    ans = len(np.unique(tads[idx, :2]))
    return ans, idx

if __name__ == '__main__':
    pass



import os
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from data_process import load_mat, get_shuffle_mat
from get_bds import get_bds, select_bds
from dp import get_dp_mat, get_tads
from quant_reg import quant_reg
import numpy as np
import time

def ks_h(mat, bd, min_sz=7, max_sz=200, p=1, threshold=0.1, mean_threshold = 0.1):
    if bd == 0 or bd > len(mat) - min_sz - 2:
        return []
    ks_stat = []
    ans1 = (mat[bd, (bd+1):min(len(mat)-1, bd+max_sz)] - mat[bd-1, (bd-1+1):min(len(mat)-2, bd-1+max_sz)]) 
    ans2 = (mat[bd, (bd+1):min(len(mat)-1, bd+max_sz)] - mat[bd+1, (bd+1+1):min(len(mat), bd+1+max_sz)]) 
    ans = np.array([ans1[i]+ans2[i] if ans1[i]>threshold and ans2[i]>threshold else 0 for i in range(len(ans1))])
    if np.sum(ans>0) < 20:
        return []
    NR, Nmiss = 0, 0
    for i in range(len(ans)):
        NR += ans[i] ** p if ans[i]>0 else 0
        Nmiss += 1 if ans[i]<=0 else 0
    if NR == 0:
        return []
    hit, miss = 0, 0
    for i in range(len(ans)):
        if i < min_sz-3:
            hit += ans[i] ** p if ans[i]>0 else 0
            miss += 1 if ans[i]<=0 else 0
        else:
            hit += ans[i] ** p if ans[i]>0 else 0
            miss += 1 if ans[i]<=0 else 0
            try:
                ks_stat.append(hit/NR-miss/Nmiss)
            except:
                ks_stat.append(hit/NR)
    max_idx = np.argmax(ks_stat)
    if max_idx == 0 or np.mean(ks_stat[:(max_idx+1)]) < mean_threshold:
        return []
    else:
        return [[bd], [bd, bd+min_sz+max_idx-2]]
def ks_v(mat, bd, min_sz=7, max_sz=200, p=1, threshold=0.1, mean_threshold = 0.1):
    if bd == len(mat)-1 or bd < min_sz + 1:
        return []
    ks_stat = []
    ans1 = (mat[max(1, bd-max_sz+1):bd, bd] - mat[max(0, bd-1-max_sz+1):bd-1, bd-1]) 
    ans2 = (mat[max(1, bd-max_sz+1):bd, bd] - mat[max(2, bd+1-max_sz+1):bd+1, bd+1]) 
    ans = np.array([ans1[i]+ans2[i] if ans1[i]>threshold and ans2[i]>threshold else 0 for i in range(len(ans1))])
    if np.sum(ans>0) < 20:
        return []
    NR, Nmiss = 0, 0
    for i in range(len(ans)):
        NR += ans[i] ** p if ans[i]>0 else 0
        Nmiss += 1 if ans[i]<=0 else 0
    if NR == 0:
        return []
    hit, miss = 0, 0
    for i in range(len(ans)):
        if i < min_sz-3:
            hit += ans[-(i+1)] ** p if ans[-(i+1)]>0 else 0
            miss += 1 if ans[-(i+1)]<=0 else 0
        else:
            hit += ans[-(i+1)] ** p if ans[-(i+1)]>0 else 0
            miss += 1 if ans[-(i+1)]<=0 else 0
            try:
                ks_stat.append(hit/NR-miss/Nmiss)
            except:
                ks_stat.append(hit/NR)
    max_idx = np.argmax(ks_stat)
    if max_idx == 0 or np.mean(ks_stat[:(max_idx+1)]) < mean_threshold:
        return []
    else:
        return [[bd-min_sz-max_idx+2, bd], [bd]]
def _get_stripe(mat, bds, min_sz=7, max_sz=200, p=1, threshold=0.1, mean_threshold=0.1):
    stripe = []
    for bd in bds:
        stripe_h = ks_h(mat, int(bd), min_sz, max_sz, p, threshold, mean_threshold)
        stripe_v = ks_v(mat, int(bd), min_sz, max_sz, p, threshold, mean_threshold)
        if len(stripe_h) == 2:
            stripe.append(stripe_h)
        if len(stripe_v) == 2:
            stripe.append(stripe_v)
    return stripe

def get_stripe(file_name, min_sz=7, max_sz=200, p=1, threshold=0.1, mean_threshold=0.1):
    mat, cum_mat = load_mat(file_name, max_sz)
    
    bds = np.arange(len(mat))
    
    stripes = _get_stripe(mat, bds, min_sz, max_sz, p, threshold, mean_threshold)
    
    return stripes

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Jointly identify nested TADs.')
    parser.add_argument('-F', '--file_name', type=str)
    parser.add_argument('-O', '--output', type=str)
    parser.add_argument('-C', '--chr', type=str)
    parser.add_argument('-MAXSZ', '--max_sz', type=int, default=200,
                        help='Maximum size of TADs')
    parser.add_argument('-MINSZ', '--min_sz', type=int, default=7,
                        help='Minimum size of TADs')
    parser.add_argument('-P', '--p', type=float, default=1)
    parser.add_argument('-T', '--threshold', type=float, default=0.1)
    args = parser.parse_args()
    stripes = get_stripe(args.file_name, max_sz=args.max_sz, min_sz=args.min_sz, p=args.p, threshold=args.threshold, mean_threshold=args.threshold)

    res = []
    for stripe in stripes:
        if len(stripe[0]) == 1:
            res.append(['chr'+args.chr, stripe[0][0], stripe[0][0], 'chr'+args.chr, stripe[1][0], stripe[1][1]])
        else:
            res.append(['chr'+args.chr, stripe[0][0], stripe[0][1], 'chr'+args.chr, stripe[1][0], stripe[1][0]])
    np.savetxt(args.output, res, fmt='%s', delimiter='\t')



import numpy as np
import numba
import pandas as pd

@numba.jit(nopython=True)
def cum_sum(mat):
    # cum_mat[i, j] = sum(mat[:(i+1), :(j+1)])
    cum_mat = mat.copy()
    for i in range(mat.shape[0]):
        for j in range(1, mat.shape[0]):
            cum_mat[i, j] += cum_mat[i, j-1]
        if i > 0:
            for j in range(mat.shape[0]):
                cum_mat[i, j] += cum_mat[i-1, j]
    return cum_mat

@numba.jit(nopython=True)
def normalization(mat, max_sz):
    # Normalize the diagonals [0, 2*max_sz-1]
    for i in range(2*max_sz):
        diag = np.diag(mat, i)
        diag_min = np.min(diag)
        diag_max = np.percentile(diag, 99.9)
        if diag_max == 0 or diag_max == diag_min:
            continue
        for j in range(mat.shape[0]-i):
            mat[j, j+i] = mat[j+i, j] = min(1, (diag[j] - diag_min) / \
                                        (diag_max - diag_min))
    return mat

def _load_mat(file_name, max_sz):
    # load the contact matrix
    f = open(file_name, 'r')
    line = f.readline()
    if len(line.split()) == 1:
        with open(file_name, 'r') as f:
            for i, line in enumerate(f):
                line = line.split(',')
                if i == 0:
                    mat_size = len(line)
                    mat = np.zeros((mat_size, mat_size))
                for j in range(max(0, i-2*max_sz+1), min(mat_size, i+2*max_sz)):
                    mat[i, j] = float(line[j])
    else:
        with open(file_name, 'r') as f:
            for i, line in enumerate(f):
                line = line.split()
                if i == 0:
                    mat_size = len(line)
                    mat = np.zeros((mat_size, mat_size))
                for j in range(max(0, i-2*max_sz+1), min(mat_size, i+2*max_sz)):
                    mat[i, j] = float(line[j])
    return mat

def load_mat(file_name, max_sz):
    # load and cum_sum matrix of file_name
    mat = _load_mat(file_name, max_sz)
    mat = normalization(mat, max_sz)
    cum_mat = cum_sum(mat)
    return mat, cum_mat

def shuffle(mat, max_sz):
    # shuffle mat at each geometric distance within 2*max_sz
    mat_shuffle = mat.copy()
    l = mat_shuffle.shape[0]
    for i in range(2*max_sz):
        diag = np.diag(mat_shuffle, i)
        diag.setflags(write=True)
        np.random.shuffle(diag)
        #np.random.shuffle(np.diag(mat_shuffle, i))
    for i in range(l):
        for j in range(i+1, min(l, i+2*max_sz)):
            mat_shuffle[j, i] = mat_shuffle[i, j]
    return mat_shuffle

def get_shuffle_mat(mat, max_sz):
    # return shuffled mat and its cum_sum matrix
    mat_shuffle = shuffle(mat, max_sz)
    cum_mat_shuffle = cum_sum(mat_shuffle)
    return mat_shuffle, cum_mat_shuffle



from setuptools import setup, find_packages

setup(
    name='JOnTADS',
    version='0.5',
    author='Qiuhai Zeng',
    author_email='qiuhai.stat@gmail.com',
    description='JOnTADS: a unified caller for TADs and stripes in Hi-C data',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.10.5',
    install_requires=[
        'numba==0.56.4', 
        'numpy==1.23.5',
        'scipy==1.12.0',
        'qpsolvers==2.7.3'
    ],
    entry_points={
        'console_scripts': [
            'JOnTADS=JOnTADS.JOnTADS:main',  # "script_name=package.module:function"
        ],
    },
    url='https://github.com/qunhualilab/JOnTADS',
)



import numpy as np
import multiprocessing

def maximize_sum_ft(bds, ft, interval, pos):
    # pos: points that must be included
    p = 0
    res = [[[-interval], -1]] # [[[bds], sum]]
    for i in range(len(bds)):
        if p < len(pos) and bds[i] == pos[p]:
            max_sum = -1
            for j in range(len(res)):
                res_cur = res[-(j+1)]
                bds_cur = res_cur[0]
                sum_cur = res_cur[1]
                if bds[i] - bds_cur[-1] + 1 < interval:
                    if sum_cur + ft[bds[i]] - ft[bds_cur[-1]] > max_sum:
                        max_sum = sum_cur + ft[bds[i]] - ft[bds_cur[-1]]
                        bds_next = bds_cur[:-1] + [bds[i]]
                else:
                    if ft[bds[i]] + sum_cur > max_sum:
                        bds_next = bds_cur + [bds[i]]
                        max_sum = ft[bds[i]] + sum_cur
            p += 1
            res.append([bds_next, max_sum])
            res = [res[-1]]
        else:
            bds_next = res[-1][0]
            max_sum = res[-1][1]
            for j in range(len(res)):
                res_cur = res[-(j+1)]
                bds_cur = res_cur[0]
                sum_cur = res_cur[1]
                if bds[i] - bds_cur[-1] + 1 < interval:
                    if ft[bds[i]] > ft[bds_cur[-1]]:
                        if sum_cur + ft[bds[i]] - ft[bds_cur[-1]] > max_sum:
                            max_sum = sum_cur + ft[bds[i]] - ft[bds_cur[-1]]
                            bds_next = bds_cur[:-1] + [bds[i]]
                else:
                    if ft[bds[i]] + sum_cur > max_sum:
                        bds_next = bds_cur + [bds[i]]
                        max_sum = ft[bds[i]] + sum_cur
            res.append([bds_next, max_sum])
            res = res[-interval+1:]
    return res[-1][0][1:]
    
def _select_bds(bds, ft, interval):
    pos = bds[np.where(ft[bds] == np.max(ft[bds]))[0]]
    pos = maximize_sum_ft(pos, ft, interval, [-1])
    bds = maximize_sum_ft(bds, ft, interval, pos)
    return bds    

def select_bds(bds, ft, interval):
    bds_selected = []
    n = len(bds)
    i = 0
    while i <= n-1:
        bds_part = [bds[i]]
        for j in range(1, n-i+1):
            if i+j <= n-1 and bds[i+j] - bds_part[-1] + 1 < interval:
                bds_part.append(bds[i+j])
            else:
                i += j
                break
        bds_selected += _select_bds(np.array(bds_part), ft, interval)
    return bds_selected

def get_diag_ft_diff(mat, max_sz, min_sz, num):
    # get feature difference of diagonal elements
    cut_p = np.linspace(min_sz, max_sz, num, dtype=int)
    diag_h, diag_v = [[] for i in range(num)], [[] for i in range(num)]
    diag_h_diff, diag_v_diff = [[] for i in range(num)], [[] for i in range(num)]                    
    for i in range(num):
        for j in range(mat.shape[0]):
            if j <= mat.shape[0] - cut_p[i]:
                diag_h[i].append(np.mean(np.sort(mat[j, j+2 : j+cut_p[i]])[-min_sz+2:]))
            if j >= cut_p[i]-1:
                diag_v[i].append(np.mean(np.sort(mat[j-cut_p[i]+1 : j-1, j])[-min_sz+2:]))
    
    for i in range(num):
        diag_h[i] = np.array(diag_h[i])
        diag_v[i] = np.array(diag_v[i][::-1])
    
    for i in range(num):
        diag_h_diff[i] = np.diff(diag_h[i])
        diag_v_diff[i] = np.diff(diag_v[i])
        
    return diag_h_diff, diag_v_diff

from scipy.stats import norm
def norm_ref(c):
    return 1-norm.cdf(norm.ppf(0.9)+c)
def cur_prop(diff, c):
    return np.sum(diff > np.percentile(diff, 90) + c * np.std(diff))/len(diff)
def bisect_trsd(diff):
    left = 1
    right = (np.max(diff) - np.percentile(diff, 90))/np.std(diff)
    mid = (left+right)/2
    for _ in range(10):
        mid = (left+right)/2
        if norm_ref(mid)/cur_prop(diff,mid) < 0.2:
            right = mid
        elif norm_ref(mid)/cur_prop(diff,mid) > 0.2:
            left = mid
        else:
            return mid
    return mid
from scipy.stats import norm
def gmm(data):
    tau, sigma1, sigma2 = 0.5, np.std(data), np.std(data)/100
    Qold = -np.inf
    T = np.ones(len(data))*0.5
    Qnew = 0
    for i in range(len(data)):
        Qnew += T[i] * (np.log(tau)-np.log(sigma1)-1/2*data[i]**2/sigma1**2-1/2*np.log(2*np.pi)) + (1-T[i]) * (np.log(1-tau)-np.log(sigma2)-1/2*data[i]**2/sigma2**2-1/2*np.log(2*np.pi))
    while Qnew-Qold > 0.001:
        Qold = Qnew
        T = np.zeros(len(data))
        for i in range(len(data)):
            T[i] = tau*norm.pdf(data[i], loc=0, scale = sigma1) / (tau*norm.pdf(data[i], loc=0, scale = sigma1) + (1-tau)*norm.pdf(data[i], loc=0, scale = sigma2))
        Qnew = 0
        for i in range(len(data)):
            Qnew += T[i] * (np.log(tau)-np.log(sigma1)-1/2*data[i]**2/sigma1**2-1/2*np.log(2*np.pi)) + (1-T[i]) * (np.log(1-tau)-np.log(sigma2)-1/2*data[i]**2/sigma2**2-1/2*np.log(2*np.pi))
        tau = np.mean(T)
        sigma1 = np.sqrt(np.sum(T*data**2)/np.sum(T))
        sigma2 = np.max([np.sqrt(np.sum((1-T)*data**2)/np.sum(1-T)), sigma1/1000])
    return tau, sigma1, sigma2

from scipy.optimize import minimize
from scipy.stats import norm
def optim(data):
    bnds = ((0, 1), (1e-10, None), (1e-10, None))
    def llh(pars, data):
        tau, sigma1, sigma2 = pars[0], pars[1], pars[2]
        return -np.sum(np.log(tau*norm.pdf(data, loc=0, scale=sigma1)+(1-tau)*norm.pdf(data, loc=0, scale=sigma2)+1e-10))
    pars = np.array([0.5, np.std(data), np.std(data)/10000])
    res = minimize(llh, pars, args=data, bounds=bnds, tol=1e-6, method='L-BFGS-B')
    return res.x

def _get_bds(diag_ft_diff, diag_ft_diff_shuffle, percentile):
    tau, sigma1, sigma2 = optim(diag_ft_diff)
    if sigma2 > np.max(diag_ft_diff):
        tau, sigma2 = 1, 0
    if sigma1 > np.max(diag_ft_diff):
        tau, sigma1, sigma2 = 1, sigma2, 0
    if sigma2 > sigma1:
        tau, sigma1, sigma2 = 1-tau, sigma2, sigma1
    return np.where(diag_ft_diff >= (2*sigma1+(1-tau)*sigma2))[0] + 1

def para_bds(diag_diff, ft, interval, percentile):
    cores = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(processes=cores)
    tasks = [(diag_diff[i], diag_diff[i], percentile) for i in range(len(diag_diff))]
    res = pool.starmap(_get_bds, tasks)
    pool.close()
    pool.join()
    return res
    
def get_bds(mat, mat_shuffle, max_sz, min_sz, percentile, num):
    ft = []
    for i in range(mat.shape[0]):
        ft.append(np.mean(np.sort(mat[i, max(0, i-max_sz+1):min(mat.shape[0], i+max_sz)])[-min_sz:]))
    ft = np.array(ft)
    
    n = mat.shape[0]
    diag_h_diff, diag_v_diff = get_diag_ft_diff(mat, max_sz, min_sz, num)
    diag_h_diff_shuffle, diag_v_diff_shuffle = get_diag_ft_diff(mat_shuffle, max_sz, min_sz, num)

    interval = int((min_sz+1)/2)
    bds_h, bds_v = [], []
    resh = para_bds(diag_h_diff, ft, interval, percentile)
    for _bds_h in resh:
        bds_h += select_bds(_bds_h, ft, interval)
    resv = para_bds(diag_v_diff, ft, interval, percentile)
    for _bds_v in resv:
        bds_v += (n-1- np.array(select_bds(_bds_v, ft[::-1], interval))).tolist()
    #for i in range(len(diag_h_diff)):
        #_bds_h = _get_bds(diag_h_diff[i], diag_h_diff_shuffle[i], percentile)
        #_bds_h = select_bds(_bds_h, ft, interval)
        #bds_h += _bds_h
        #_bds_v = _get_bds(diag_v_diff[i], diag_v_diff_shuffle[i], percentile)
        #_bds_v = (n-1- np.array(select_bds(_bds_v, ft[::-1], interval))).tolist()
        #bds_v += _bds_v
        
    bds = np.unique(bds_h+bds_v)
    return select_bds(bds, ft, interval)

if __name__ == '__main__':
    pass



import numba
import pandas as pd
import numpy as np

@numba.njit
def get_dp_mat(bds, mat, cum_mat, max_sz, min_sz, penalty):
    # 0: intermediate point position
    # 1: outer TAD whether included
    # 2: left boundary whether included
    # 3: right boundary whether included
    # 4: sum of entry values
    # 5: num of points used
    # 6: sum of TAD scores
    # 7: TAD score
    # 8: the maximum TAD size of current hierarchy
    l = len(bds)
    dp_mat = []
    for i in range(l):
        dp_mat.append([[-1, 0, 0, 0, 0., 0, 0., 0., 0]
                       for _ in range(l)])
    for i in range(1, int(np.ceil((max_sz-1)/np.ceil((min_sz-1)/2))+1)+1):
        for j in range(len(bds)-i):
            if bds[i + j] - bds[j] + 1 < min_sz:
                # TAD size >= min_sz
                continue
            if bds[i + j] - bds[j] + 1 <= max_sz:
                # TAD size <= max_sz
                tad_size = bds[i + j] - bds[j] + 1
                sub_mat_sum, sub_mat_num_points = sum_sub_mat(bds[j], bds[j + i], cum_mat)
                
                outer_include = 1
                corner_sum, corner_num = corner(bds[j], bds[j + i], max(int((min_sz+1)/2), int((bds[j+i]-bds[j]+1)/10)), cum_mat)
                if (corner_sum * sub_mat_num_points) < 0.6 * (sub_mat_sum * corner_num):
                    outer_include = 0
                
                sub_mat_surr_width = int(round(sub_mat_num_points /
                                       (2. * (tad_size-1))))
                sub_mat_surr_mean = cal_surr_mean(bds[j], bds[i + j],
                                          cum_mat, sub_mat_surr_width)
                # outer TAD score
                size = int(bds[j+i] - bds[j] + 1)
                score = 0
                score_max = max(0, sub_mat_sum / sub_mat_num_points - sub_mat_surr_mean - penalty[size-min_sz])
                
                inter = -1
                for k in range(j+1, j+i):
                    # nested TADs should be at least min_sz-1 smaller
                    if dp_mat[j][k][8] > tad_size - min_sz + 1 or dp_mat[k][j+i][8] > tad_size - min_sz + 1 or \
                    tad_size < 1.1 * max(dp_mat[j][k][8], dp_mat[k][j+i][8]):
                    #if dp_mat[j][k][8] > tad_size - min_sz + 1 or dp_mat[k][j+i][8] > tad_size - min_sz + 1:
                        if dp_mat[j][k][6] + dp_mat[k][j+i][6] > score_max:
                            score_max = dp_mat[j][k][6] + dp_mat[k][j+i][6]
                            score = 0
                            inter = k
                        continue

                    sub_mat_sum_left = sub_mat_sum
                    sub_mat_num_points_left = sub_mat_num_points
                    if dp_mat[j][k][4] != 0:
                        sub_mat_sum_left -= dp_mat[j][k][4]
                        sub_mat_num_points_left -= dp_mat[j][k][5]
                    if dp_mat[k][j+i][4] != 0:
                        sub_mat_sum_left -= dp_mat[k][j+i][4]
                        sub_mat_num_points_left -= dp_mat[k][j+i][5]
                    if dp_mat[j][k][3] != 0 and dp_mat[k][j+i][2] != 0:
                        sub_mat_sum_left += mat[bds[k], bds[k]]
                        sub_mat_num_points_left += 1
                    sub_mat_mean_left = sub_mat_sum_left / \
                                        sub_mat_num_points_left

                    sub_mat_surr_width_left = int(round(
                        sub_mat_num_points_left / (2. * (tad_size-1))))
                    #sub_mat_surr_width_left = max(int(round(
                    #    sub_mat_num_points_left / (2. * (tad_size-1)))), min_sz-1)
                    #sub_mat_surr_width_left = tad_size - 1
                    sub_mat_surr_mean_left = cal_surr_mean(
                        bds[j], bds[j+i], cum_mat, sub_mat_surr_width_left)
                    if outer_include == 1 and sub_mat_mean_left - sub_mat_surr_mean_left - penalty[size-min_sz] > 0:
                        if sub_mat_mean_left - sub_mat_surr_mean_left + \
                                dp_mat[k][j+i][6] + dp_mat[j][k][6] - \
                                penalty[size-min_sz] > score_max:
                            score_max = sub_mat_mean_left - \
                                        sub_mat_surr_mean_left + \
                                        dp_mat[j][k][6] + dp_mat[k][j+i][6] - penalty[size-min_sz]
                            score = sub_mat_mean_left - \
                                    sub_mat_surr_mean_left - penalty[size-min_sz]
                            inter = k
                    else:
                        if dp_mat[j][k][6] + dp_mat[k][j+i][6] > score_max:
                            score_max = dp_mat[j][k][6] + dp_mat[k][j+i][6]
                            score = 0
                            inter = k
                            
                if score > 0:
                    dp_mat[j][j+i][0] = inter
                    dp_mat[j][j+i][1] = 1
                    dp_mat[j][j+i][2] = 1
                    dp_mat[j][j+i][3] = 1
                    dp_mat[j][j+i][4] = sub_mat_sum
                    dp_mat[j][j+i][5] = sub_mat_num_points
                    dp_mat[j][j+i][6] = score_max
                    dp_mat[j][j+i][7] = score
                    dp_mat[j][j+i][8] = bds[j+i] - bds[j] + 1
                elif inter != -1:
                    dp_mat[j][j+i][0] = inter
                    dp_mat[j][j+i][1] = 0
                    if dp_mat[j][inter][2] == 1:
                        dp_mat[j][j+i][2] = 1
                    if dp_mat[inter][j+i][3] == 1:
                        dp_mat[j][j+i][3] = 1
                    if dp_mat[j][inter][3] == 1 and \
                            dp_mat[inter][j+i][2] == 1:
                        dp_mat[j][j+i][4] = dp_mat[j][inter][4] + \
                                            dp_mat[inter][j+i][4] - \
                                            mat[bds[inter], bds[inter]]
                        dp_mat[j][j+i][5] = dp_mat[j][inter][5] + \
                                            dp_mat[inter][j+i][5] - 1
                    else:
                        dp_mat[j][j+i][4] = dp_mat[j][inter][4] + \
                                            dp_mat[inter][j+i][4]
                        dp_mat[j][j+i][5] = dp_mat[j][inter][5] + \
                                            dp_mat[inter][j+i][5]
                    dp_mat[j][j+i][6] = dp_mat[j][inter][6] + \
                                        dp_mat[inter][j+i][6]
                    dp_mat[j][j+i][8] = max(dp_mat[j][inter][8],
                                            dp_mat[inter][j+i][8])
                elif outer_include == 1 and score_max > 0:
                    dp_mat[j][j+i][0] = inter
                    dp_mat[j][j+i][1] = 1
                    dp_mat[j][j+i][2] = 1
                    dp_mat[j][j+i][3] = 1
                    dp_mat[j][j+i][4] = sub_mat_sum
                    dp_mat[j][j+i][5] = sub_mat_num_points
                    dp_mat[j][j+i][6] = score_max
                    dp_mat[j][j+i][7] = score_max
                    dp_mat[j][j+i][8] = bds[j+i] - bds[j] + 1
                else:
                    continue
            else:
                continue
    
    return dp_mat

@numba.njit
def corner(bd1, bd2, interval, cum_mat):
    sub_mat_sum = cum_mat[bd1+interval-1, bd2] - cum_mat[bd1+interval-1, bd2-interval] - cum_mat[bd1-1, bd2] + cum_mat[bd1-1, bd2-interval]
    return sub_mat_sum, interval**2

@numba.njit
def sum_sub_mat(l_b, r_b, cum_mat):
    # l_b: left boundary
    # r_b: right boundary
    sub_mat_size = r_b - l_b + 1

    if l_b == 0:
        return cum_mat[r_b, r_b], sub_mat_size**2

    sub_mat_sum = cum_mat[r_b, r_b] - cum_mat[r_b, l_b-1] - \
                  cum_mat[l_b-1, r_b] + cum_mat[l_b-1, l_b-1]
    return sub_mat_sum, sub_mat_size**2

@numba.njit
def cal_surr_mean(l_b, r_b, cum_mat, surr_width):
    # up_mean: [max(0, l_b-surr_width) : l_b-1, l_b+1 : r_b]
    # ri_mean: [l_b : r_b-1, r_b+1 : min(r_b, size(cum_mat)-1)]
    if l_b == 0:
        up_mean = 0
    else:
        up_b = max(0, l_b-surr_width)
        if up_b == 0:
            up_sum = cum_mat[l_b-1, r_b] - cum_mat[l_b-1, l_b]
        else:
            up_sum = cum_mat[l_b-1, r_b] - cum_mat[l_b-1, l_b] - \
                  cum_mat[up_b-1, r_b] + cum_mat[up_b-1, l_b]
        up_mean = up_sum / ((r_b-l_b)*(l_b-up_b))

    if r_b == cum_mat.shape[0]-1:
        ri_mean = 0
    else:
        ri_b = min(cum_mat.shape[0]-1, r_b+surr_width)
        if l_b == 0:
            ri_sum = cum_mat[r_b-1, ri_b] - cum_mat[r_b-1, r_b]
        else:
            ri_sum = cum_mat[r_b-1, ri_b] - cum_mat[r_b-1, r_b] - \
                     cum_mat[l_b-1, ri_b] + cum_mat[l_b-1, r_b]
        ri_mean = ri_sum / ((r_b-l_b)*(ri_b-r_b))
    return max(up_mean, ri_mean)

def get_tads(bds, dp_mat, max_sz, min_sz):
    # tads: [[position, [[tad], [tad]], max_score]]
    # ensure the 'inter' is an integer
    for i in range(len(bds)):
        for j in range(len(bds)):
            dp_mat[i][j][0] = int(dp_mat[i][j][0])
            
    tads = [[0, [], 0]]
    for i, bd, in enumerate(bds[1:]):
        # TAD should be smaller than (or equal to) max_sz
        if bd - bds[tads[-1][0]] + 1 > max_sz:
            tads = [tads[-1]]
            tads[-1][0] = i + 1
        else:
            tads_last = tads[-1][1]  # [[tad], [tad]]
            max_score_last = tads[-1][2]  # float

            # tads_cur: [position, [[tad], [tad]], max_score]
            for j, tads_cur in enumerate(tads[::-1]):
                if bd - bds[tads_cur[0]] + 1 > max_sz:
                    tads = tads[-j:]
                    break

                # TAD should be larger than (or equal to) mim_sz
                if bd - bds[tads_cur[0]] + 1 < min_sz:
                    continue
                else:
                    max_score = tads_cur[2] + dp_mat[tads_cur[0]][i+1][6]
                    if max_score > max_score_last:
                        tads_last = tads_cur[1] + sub_get_tads(
                            dp_mat, tads_cur[0], i+1, bds, [])
                        max_score_last = max_score
            tads.append([i+1, tads_last, max_score_last])

    return np.array(tads[-1][1])

def sub_get_tads(dp_mat, l_b, r_b, bds, tads):
    # tads: [[l_b, r_b, size, score]]
    # in order: left to right, outer to inner
    if dp_mat[l_b][r_b][1] == 1:
        tads.append([bds[l_b], bds[r_b], dp_mat[l_b][r_b][7], bds[r_b]-bds[l_b]+1])
    if dp_mat[l_b][r_b][0] != -1:
        tads = sub_get_tads(dp_mat, l_b, dp_mat[l_b][r_b][0], bds, tads)
        tads = sub_get_tads(dp_mat, dp_mat[l_b][r_b][0], r_b, bds, tads)
    return tads
